"""
 ***************************************************************************
                             Database module
                             -------------------
    begin                : Sat Dec 14 2002
    copyright            : (C) 2002 by Ricardo Niederberger Cabral
    email                : nieder@mail.ru
    Time-stamp:            <2003-01-08 02:27:56 rnc>
 ***************************************************************************

 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
"""
try:
    import sys,os,traceback,md5,random
    import marshal                          # for loading and saving img databases
    from string import *
    from Haar import *
    import EXIF
except:
    traceback.print_exc()
    print "Error importing the necessary python modules. Unable to continue."
    sys.exit()
try:
    from qt import *
except:
    Error.qtImportError()
####### DEFINES
supported_ext=["jpeg","jpe","jpg","gif","bmp","png","xbm","pnm"]       # thats what QT seems to support. gif and jpg may not be present. TODO: detect it
exif_ext=["jpeg","jpe","jpg"]
####### GLBOAL VARS
bin=[]

class ImgDB:
    def __init__(self,ndbname="",chgCb=None):
        """ndbname is the filename for a database to open on object creation
        chgCb is a callback function invoked on every db change
        """
        self.dbversion=3                # saved on the beginning of marshalled db files. For version db detection and handling
        self.fname=ndbname              # the filename of the currently opened database
        #### image metadata:
        self.imageParms=["Description","Dimensions","Filesize","Filename","Format"]
        self.blankMetaDict={}           # blank dictionary with available keys. To be copied later to every new db file
        for par in self.imageParms:
            self.blankMetaDict[par]=None        
        self.reset()
        #### Thumbnail stuff
        self.thdir=os.path.expanduser(os.path.join("~",".thumbnails","")) # path for storing thumbnails. Calculate it only once here.
        self.thdir2=os.path.expanduser(os.path.join("~",".thumbnails","normal","")) # path for storing thumbnails. Calculate it only once here.
        # just to make sure the thumbnail dir exists, so I dont need to be checking it over and over on self.getThumb()
        if not os.path.exists(self.thdir):
            os.mkdir(self.thdir)
        if not os.path.exists(self.thdir2):
            os.mkdir(self.thdir2)        
        #### end
        self.wnd=None
        self.app=None
        self.changeCallback=chgCb
        self.eng=HaarEngine(self)
        if len(ndbname):
            self.opendb()
        else:            
            self.reset()
        
    def changedDB(self):
        """helper function, call it whenever something on db is changed and the GUI should know of """
        self.dirty=1
        if self.changeCallback: self.changeCallback()        
    def remove_dead_volume(self,volumeid):
        """removes any non existant file on volume id """
        for didx in self.volumes[volumeid]:
            self.remove_dead_dir(didx)
        self.changedDB()
    def remove_dead(self):
        """removes any non existant file on this dir id """
        #TODO: add check for infinite loop. a child dir cannot reference its ancestors
        for fidx in self.img.keys():
            if not os.path.exists(self.img[fidx][0]):
                self.removeFile(fidx)
    def remove_dead_dir(self,dirid):
        """removes any non existant file on this dir id """
        #TODO: add check for infinite loop. a child dir cannot reference its ancestors
        for fidx in self.dirs[dirid][2]:
            self.remove_dead_dir(fidx)
        for fidx in self.dirs[dirid][3]:            
            if not os.path.exists(self.img[fidx][0]):
                print "Removed from database the dead file: "+self.img[fidx][0]
                self.zombie.append(fidx)
    def rescan(self):
        """recalculate coefficients for all images on the database and regenerate the db buckets  """
        #TODO: store filesize to determine if a file should be rescanned or not ?
        #this feature doesn't seem to be needed, as its very unlikely that images will be changed, only removed.
        pass
    def reset(self):
        """ reset current db, should be called on File-> New """
        ## reseting all data structures
        self.img={}                     # key is id, value is [path,parent dir id,avgl]
        self.meta={}                    # key is id, value is dictionary of meta values
        self.buckets=[]                 # coef buckets. This is a 128x128 list of [color][positive][negative] lists of ids
        self.dirs={}                    # key id, value is [0path,1-parent dir id,2-child dir ids,3-child files ids,4-child imgs filenames,5-description,6-volumeid]
        self.volumes={1:[[],"/","Default volume","Local filesystem"]}                 # key id, value is [0-list of dir ids on this volume,1-base path,2-description,3-name]
        self.zombies=[]                  # list of files removed from database and waiting a cleanup to iterate all data structures and remove references
        self.groups={}# key is group id, value is [0-name,1-description,2-parent id, 3-child imgs id, 4-child groups 5-id]
        self.opened=0                   # db has been succesfully opened
        self.dirty=0                    # has db been changed since last change ?
        self.openGroups=[]              # group ids with listvierw item expanded (so ui will set the listview state when it changed)
        self.simgroups={}# key is id value is [0-group name,1-main img id,2-list of similar images ids]
        self.metafields=self.imageParms[:] # history of entered metafields        
        self.textqueryhistory=[]        # key is text str, value is list of [field,value] pairs
        for i in range(16384):
            self.buckets.append([[[],[]],[[],[]],[[],[]]])  # color[positive,negative]
        # standard group where all images are added to if nothing else is specified
        self.groups[1]=["Orphan","Images which haven't been assigned to any other group.",-1,[],[]]
    def opendb(self):        
        print "Opening database file "+self.fname+" ..."
        rdbversion=1                    # database version read from db file
        try:
            self.kbsize=os.stat(self.fname).st_size
        except:
            self.kbsize=0
        try:
            f=open(self.fname,"rb")
            rdbversion=marshal.load(f)
            if rdbversion < 3:
                self.v1files=marshal.load(f)
                self.v1buckets=marshal.load(f)
                self.v1dirs=marshal.load(f)
                self.v1dirtree=marshal.load(f)
                self.v1groups=marshal.load(f)            
                if rdbversion>1:
                    self.v1meta=marshal.load(f)
            else: # load db > v3:
                self.img=marshal.load(f)
                self.meta=marshal.load(f)                
                self.buckets=marshal.load(f)
                self.dirs=marshal.load(f)
                self.volumes=marshal.load(f)
                self.zombies=marshal.load(f)
                self.groups=marshal.load(f)                
                self.simgroups=marshal.load(f)
                self.metafields=marshal.load(f)
                self.textqueryhistory=marshal.load(f)                                
                self.openGroups=marshal.load(f)                                
            f.close()
            self.opened=1            
        except:
            print "Error opening database, starting with an empty one."
            self.reset()
            return 0
        if rdbversion==1 or rdbversion==2:
            # import a dbversion 1 db and increment it to the next version so the next for catches it and do the proper imports for the next version
            print "Importing version 2 image database to version 3..."            
            for dir in self.v1dirtree.keys():
                if not self.v1dirtree[dir]: # is root dir
                    self.addDir(1,dir,1,[],1)
            # import descriptions
            for k in self.v1meta.keys():
                try:
                    if self.v1meta[k]["Description"]:
                        for fid in self.meta.keys():
                            if self.meta[fid]["Filename"]==self.v1meta[k]["Filename"]:
                                self.meta[fid]["Description"]=self.v1meta[k]["Description"]
                except:
                    pass
            self.dirty=1
        print "Done."
        if self.changeCallback: self.changeCallback()        
        return 1
    def changefname(self,nname):
        self.fname=nname
        self.dirty=1
        if self.changeCallback: self.changeCallback()
    def generateImageID(self):
        """ need to return something unique"""
        return random.randrange(999999999)
    def savedb(self):
        try:
            f=open(self.fname,"wb")
            marshal.dump(self.dbversion,f)
            #marshal.dump(1,f)
            marshal.dump(self.img,f)
            marshal.dump(self.meta,f)
            marshal.dump(self.buckets,f)
            marshal.dump(self.dirs,f)
            marshal.dump(self.volumes,f)
            marshal.dump(self.zombies,f)
            marshal.dump(self.groups,f)
            marshal.dump(self.simgroups,f)            
            marshal.dump(self.metafields,f)            
            marshal.dump(self.textqueryhistory,f)            
            marshal.dump(self.openGroups,f)            
            f.close()
        except:
            traceback.print_exc()            
            print "Error saving database."
            return 0
        self.dirty=0
        return 1
        print "Saved database file: ",self.fname
    def getImageFormat(self,fname):
        #TODO: finish
        return "JPEG"
    def extIsImg(self,file):
        """ tests if this file has a supported image format extension"""
        ext=rfind(file,".")        
        if ext==-1:return 0
        ext=lower(file[ext+1:])        
        if ext in supported_ext:  # if it is a supported image filetype
            return ext
        return None
    def updateGroups(self,app,simthresd=50):
        """ groups by similarity the imgs id listed on targetids. app is the main application (for processEvents calls). simthresd is the similarity threshold for choosing groups"""
        solved=[]
        print "Grouping with threshold of ",simthresd
        addDirProgressBar=QProgressDialog( "Creating groups, please wait...", "Abort", 10,app.mainWidget(), "addProgress", 1 )
        targetids=self.img.keys()        
        addDirProgressBar.setTotalSteps(len(targetids))
        addDirProgressBar.setAutoReset(0)
        for file in targetids:
            addDirProgressBar.setProgress(addDirProgressBar.progress()+1)
            app.processEvents()
            if addDirProgressBar.wasCancelled():
                addDirProgressBar.cancel()
                addDirProgressBar.done(1)
                addDirProgressBar.close(1)                                    
                return            
            if file in solved: continue
            solved.append(file)
            #self.simgroups={}               # key is id value is [group name,main img id,list of similar images ids]
            ngid=self.generateImageID()            
            self.simgroups[ngid]=[self.meta[file]["Filename"],file,[]]
            aImage=self.getThumb(self.img[file][0])
            if not aImage:
                print "Autogroupping error: Unable to generate thumbnail for file "+selg.img[file][0]
                continue
            results=self.eng.query(aImage,50,targetids)
            if not len(results):
                print "exception: image returned no similar img"
                continue
            for res in results:
                if (-1*res.pair[1]) > simthresd:
                    self.simgroups[ngid][2].append(res.pair[0])
                    solved.append(res.pair[0])
        #TODO: whats the right way of making sure this damn progressdialogs disappear ?
        addDirProgressBar.cancel()
        addDirProgressBar.done(1)
        addDirProgressBar.close(1)
        self.changedDB()
    def hashStr(self,str):
        return md5.new(str).hexdigest()
    def getThumb(self,fname,imgDims=None):
        """
        returns a 128x128 image of the file fname. If thummbn found, open it, otherwise create one.
        Using the standard from http://triq.net/~pearl/thumbnail-spec/       
        """   
        #TODO: not taking in consideration that fname may contain ~. I decided not to do a os.path.expanduser because that may slow down this code
        aImage=None
        thname=self.thdir2+ md5.new("file://"+fname).hexdigest()+".png"
        if not os.path.exists(thname):
            aImage=QImage()
            if not aImage.load(fname):
                print "Error loading image file while creating a thumbnail for it:" + fname
                return None
            if imgDims:
                imgDims[0]=aImage.width()
                imgDims[1]=aImage.height()
            aImage=aImage.scale(128,128)                
            aImage.save(thname,"PNG")
        if not aImage: # load it from thumbnail file
            aImage=QImage()
            if not aImage.load(thname):
                print "Error loading thumbnail file " + thname
            if (aImage.width() != 128) or (aImage.height() != 128):
                #TODO: if the thumbnail available is not 128x128 I'll be unpolite and replace it with a 128x128 one
                try:
                    os.remove(thname)
                    return self.getThumb(fname)
                except:
                    print "Error regenerating thumbnail for "+thname
                    return None
        return aImage        
    def addFile(self,fname,newid,dirid):
        """adds this filename to database. First, loading it, then calculating its haar transform
        and then adding the image index to all respective buckets.
        Each self.files list element is [filename,avg lum]
        isDir should be set to the path file if this file is being added as a batch dir
        callBack is [MainApp,QProgress]
        """
        if self.dirs.has_key(dirid):
            if fname in self.dirs[dirid][4]:
                print fname + " already on database."
                return 0
        if not fname:
            print "Attempt to open an empty file string"
            return 0
        imgDims=[0,0]
        aImage=self.getThumb(fname,imgDims)
        if not aImage:
            print "Error opening/decoding image file " + fname
            return 0        
        sig=self.eng.calcHaar(aImage,40)
        if not len(sig):
            print "Error adding file ",fname
            return 0
        data=sig[1]
        #lenf=len(self.files)
        for i in range(16384):
            for c in range(3):
                if data[i][c]>0:
                    self.buckets[i][c][0].append(newid)
                elif data[i][c]<0:
                    self.buckets[i][c][1].append(newid)
        self.img[newid][2]=sig[0]
        ## init metadata
        try:
            self.meta[newid]=self.blankMetaDict.copy()
            try:
                self.meta[newid]["Filename"]=os.path.split(fname)[-1]
                self.meta[newid]["Filesize"]=self.prettysize(os.stat(fname).st_size)
                if imgDims[0]:
                    self.meta[newid]["Dimensions"]=str(imgDims[0])+" x "+str(imgDims[1])
                self.meta[newid]["Format"]=self.getImageFormat(fname)                
            except:
                traceback.print_exc()
                print "Error setting basic metadata for "+fname
                self.meta[newid]["Filename"]=fname
        except:
            print "Error gathering metadata for ",fname
            traceback.print_exc()
        return 1
    def prettysize(self,uint):
        """returns a string after converting the integer uint and adding periods to enhance readability """
        a=str(uint)
        b=""
        for i in range(len(a)):
            if not i%3:
                b="."+b
            b=a[len(a)-1-i]+b
        return b[:-1] + " bytes"
    def addDir(self,volid,path,groupid,aborted=[],recursive=0,calledRecurs=-1,igntext="",minsize=None):
        """Add all the files on this dir.
        aborted is an empty list that should be nonempty when you want this adddir call to be aborted
        calledRecurs is the id of the dir that should be the parent to the one at path
        igntext and minsize are file criteria to ignore
        """
        if aborted: return
        print "Adding dir ",path
        startt=time.time()
        dfiles=os.listdir(path)
        if not len(dfiles):return
        if path[-1] != '/': path=path+'/'
        myLabel="Adding directory \""+path+"\" ..."
        mySteps=len(dfiles)
        myStep=0        
        hasImage=0
        addDirPB=QProgressDialog( "Adding directory...", "Abort", 10,self.wnd)
        addDirPB.setAutoReset(0)        
        addDirPB.setLabelText(myLabel)
        addDirPB.reset()
        addDirPB.setTotalSteps(mySteps)
        ndid=-1        
        for dirid in self.volumes[volid][0]:
            if self.dirs[dirid][0]==path:
                ndid=dirid                 # this dir is already on dbase
                break
        if  ndid == -1:
            #isnewdir=1
            ndid=self.generateImageID()            
        #self.dirs={}                    # key id, value is [path,parent dir id,child dir ids,child files ids,child imgs filenames,description,volumeid]
        childdirs=[]
        childfiles=[]
        childfilenames=[]
        for file in dfiles:
            addDirPB.setProgress(addDirPB.progress()+1)
            if self.app: self.app.processEvents()
            if addDirPB.wasCancelled():
                aborted.append(1)
                #self.changedDB()
                break
            if os.path.isdir(path+file) and recursive:
                if file[0]=='.': continue # do not add dirs starting with . (also excludes .thumbnail dirs)
                # update paths tree                
                myStep=addDirPB.progress()
                addDirPB.close(1)                
                thasimg=self.addDir(volid,path+file,groupid,aborted,1,ndid,igntext,minsize)
                addDirPB=QProgressDialog( "Adding directory...", "Abort", 10,self.wnd)                
                addDirPB.setLabelText(myLabel)
                addDirPB.setTotalSteps(mySteps)
                addDirPB.setProgress(myStep)                                
                addDirPB.show()                
                if thasimg == -1: continue                
                childdirs.append(thasimg)
                hasImage=1
                continue
            # is file
            ext=self.extIsImg(file)
            if ext: #its a file, now just check if its an image
                if file[0]=='.': continue # do not add imgs starting with . (also excludes .thumbnail dirs)
                if igntext and find(path+file,igntext) != -1: continue
                try:
                    if minsize and os.stat(path+file).st_size < minsize: continue                        
                except:
                    # what the hell am I supposed to do with a file that cant even be stat'd ?
                    continue
                nfid=self.generateImageID()
                self.img[nfid]=[path+file,ndid,None]                
                if self.addFile(path+file,nfid,ndid):
                    hasImage=1
                    childfiles.append(nfid)
                    childfilenames.append(path+file)
                    self.groups[groupid][3].append(nfid)
                    # extract EXIF data if is jpeg:
                    if ext in exif_ext:
                        self.extractEXIF(nfid)
                else: #error adding image
                    del self.img[nfid]
        if hasImage and not aborted: # only add to databse if has image isinde
            #TODO if the user add a given dir to 2 different volumes, that dir will b included twice. To fix it, include a loop that checks all volumes for this path
            if ndid not in self.volumes[volid][0]:self.volumes[volid][0].append(ndid)
            self.dirs[ndid]=[path,calledRecurs,childdirs,childfiles,childfilenames,"",volid]
        print "Finished adding directory."
        self.dirty=1
        #print "Adding took %f" % (time.time()-startt)
        addDirPB.close(1)
        if hasImage: return ndid
        else: return -1
    def extractEXIF(self,fid):
        try:
            file=open(self.img[fid][0], 'rb')
        except:
            print 'Unable to open file in order to extract EXIF data.'
            return
        data=EXIF.process_file(file)
        if data:
            x=data.keys()
            for i in x:
                if i in ('JPEGThumbnail', 'TIFFThumbnail'):
                    continue
                self.meta[fid][i]=data[i].printable
            if data.has_key('JPEGThumbnail'):
                self.meta[fid]['JPEGThumbnail']="True"
            else:
                self.meta[fid]['JPEGThumbnail']="False"            
    def getTotalFiles(self):
        return len(self.img.keys())
    def TextReport(self):
        ntext="Database status:\n"
        ntext=ntext+"Images: %d\n"%len(self.img.keys())
        ntext=ntext+"Volumes: %d\n"%len(self.volumes.keys())
        ntext=ntext+"Directories: %d\n"%len(self.dirs.keys())
        ntext=ntext+"Groups: %d\n"%len(self.groups.keys())
        ntext=ntext+"Similarity groups: %d\n"%len(self.simgroups.keys())
        ntext=ntext+"Metadata fields cached: %d\n"%len(self.metafields)
        ntext=ntext+"Database size: %d bytes\n"%self.kbsize
        return ntext
    def newGroup(self,text,pargid=-1):
        for grp in self.groups.keys():
            if self.groups[grp][0]==text:
                return -1
        nid=self.generateImageID()
        self.groups[nid]=[text,"",pargid,[],[]]
        if pargid != -1:
            # tell parent he has a new child
            self.groups[pargid][4].append(nid)
        self.dirty=1
        return nid
    def removeGroup(self,gid):
        if gid==1:
            print "Attempt to remove default Orphan group, aborting remove process."
            return
        # move child imgs to default orphan group (id=1)
        for id in self.groups[gid][3]:
            self.groups[1][3].append(id)
        # tell parent i died
        if self.groups[gid][2] != -1:
            try:
                self.groups[self.groups[gid][2]][4].remove(gid)
            except:
                traceback.print_exc()
                print "Error removing group from it's parent."                
        # remove children groups
        for sid in self.groups[gid][4]:
            self.removeGroup(sid)
        del self.groups[gid]
        self.dirty=1
    def removeVolume(self,gid):
        for sid in self.volumes[gid][0]:
            self.removeDir(sid)
        del self.volumes[gid]
        self.dirty=1
    def removeDir(self,gid):
        print "removing dir",self.dirs[gid][0]
        #remove child dirs
        for sid in self.dirs[gid][2]:
            self.removeDir(sid)
        #remove child imgs
        try:
            for sid in self.dirs[gid][3]:            
                self.removeFile(sid)                
        except:
            traceback.print_exc()        
        #tell parent dir im no longer a child
        try:
            if self.dirs[gid][1] != -1:
                self.dirs[self.dirs[gid][1]][2].remove(gid)
        except:
            traceback.print_exc()
        #remove from volume
        try:
            self.volumes[self.dirs[gid][6]][0].remove(gid)
        except:
            traceback.print_exc()        
        #remove me !!!        
        del self.dirs[gid]
        self.dirty=1
    def newVolume(self,txt):
        for grp in self.volumes.keys():
            if self.volumes[grp][3]==txt:
                return -1
        nid=self.generateImageID()
        while nid in self.volumes.keys():
            nid=self.generateImageID()
        self.volumes[nid]=[[],"","No description given.",txt]
    def renameVolume(self,vid,text):
        self.volumes[vid][3]=text
        self.dirty=1
    def renameGroup(self,vid,text):
        self.groups[vid][0]=text
        self.dirty=1        
    def describeVolume(self,vid,text):
        self.volumes[vid][2]=text
        self.dirty=1        
    def describeGroup(self,vid,text):
        self.groups[vid][1]=text
        self.dirty=1        
    def removeFile(self,fid):
        for did in self.dirs.keys():
            try:
                self.dirs[did][3].remove(fid)
                self.dirs[did][4].remove(self.img[fid][0])
            except:
                pass
        for did in self.groups.keys():
            try:
                self.groups[did][3].remove(fid)
            except:
                pass
        for did in self.simgroups.keys():
            try:
                self.simgroups[did][2].remove(fid)
                if self.simgroups[did][1]==fid:
                    del self.simgroups[did]
            except:
                pass
        for i in range(16384):
            for c in range(3):
                try:
                    self.buckets[i][c][0].remove(fid)
                except:
                    pass
                try:
                    self.buckets[i][c][1].remove(fid)
                except:
                    pass            
        del self.meta[fid]
        del self.img[fid]
        self.dirty=1
    def moveGroup(self,fromid,toid):
        if fromid==toid:
            return
        if toid in self.groups[fromid][4]:
            print "Invalid move. Parent -> Child"
            return
        # i guess the following commented check is not necessary
##         if self.groups[toid][2]==fromid:
##             print "Invalid move. Destination is a child of source"
##             return
        if self.groups[fromid][2] != -1:
            self.groups[self.groups[fromid][2]][4].remove(fromid)
        self.groups[fromid][2]=toid
        if toid != -1: self.groups[toid][4].append(fromid)
        self.dirty=1
    def moveFileGroup(self,fromid,fromgid,toid):
        if fromid==toid:
            return
        self.groups[fromgid][3].remove(fromid)
        self.groups[toid][3].append(fromid)
        self.dirty=1
initBin()
